/***************************************************************************
                          capture.cpp  -  description                              
                             -------------------                                         
    begin                : Thu Mar 25 14:26:46 GMT 1999
                                           
    copyright            : (C) 1999,2000 by Mike Richardson                         
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   * 
 *                                                                         *
 ***************************************************************************/


#include	<errno.h>
#include	<unistd.h>
#include	<fcntl.h>

#include	<qlist.h>
#include	<ksock.h>

#include	"ksnuffle.h"
#include	"pcaphdr.h"
#include	"netpcap.h"
#include	"capture.h"
#include	"knd_pcaploc.h"

#undef	_SHOWPKT
#undef	_PPROG

/*  PrintProg	: Print filter program for debugging purposes		*/
/*  title	: cchar *	: Title text				*/
/*  text	: cchar *	: Program text				*/
/*  prog	: BPFPROG &	: Program				*/
/*  (returns)	: void		:					*/

void	PrintProg
	(	cchar	*title,
		cchar	*text,
		BPFPROG	&prog
	)
{
#ifdef	_PPROG
	fprintf	(stderr, "Program: %s: len=%d for [%s]\n", title, prog.bf_len, text) ;

	for (uint idx = 0 ; idx < prog.bf_len ; idx += 1)
		fprintf	(stderr,
			 "%4d %4d %4d %4d\n",
			 prog.bf_insns[idx].code,
			 prog.bf_insns[idx].jt,
			 prog.bf_insns[idx].jf,
			 prog.bf_insns[idx].k) ;
#endif	/* _PPROG */
}

/*  CapCompile	: Compile filter program				*/
/*  pd		: pcap_t *	: Filter handle				*/
/*  prog	: BPFPROG *	: Program structure			*/
/*  text	: cchar *	: Program text				*/
/*  opt		: int		: Non-zero to optimise			*/
/*  mask	: ulong		: Net mask				*/
/*  (returns)	: bool		: Success				*/

bool	CapCompile
	(	pcap_t	*pd,
		BPFPROG	*prog,
		cchar	*text,
		int	opt,
		ulong	mask
	)
{
	/* Since we may compile multiple filter programs, we need to be	*/
	/* able to reset the parser, otherwise input gets left behind	*/
	/* after an error is detected. The routine "pcap_restart" does	*/
	/* this though it is not documented as part of the libpcap API	*/
	pcap_restart	(stdin) ;
	memset		(prog, 0, sizeof(prog)) ;

	return	pcap_compile (pd, prog, (char *)text, opt, mask) >= 0 ;
}


/*  CapBase								*/
/*  CapBase	: Constructor for packet capture class			*/
/*  (returns)	: CapBase	:					*/

CapBase::CapBase ()
{
	state	= MSStopped 	;
	minev	= NEVENT	;
	maxev	= -1		;
	netpcap	= NULL		;
}


/*  CapBase								*/
/*  ~CapBase	: Destructor for packet capture class			*/
/*  (returns)	:		:					*/

CapBase::~CapBase ()
{
	if (netpcap != NULL)
	{	drop_NetPCap (this, netpcap) ;
		netpcap	= NULL ;
	}
}

/*  CapBase								*/
/*  CheckStart	: Check for capture start on event trigger		*/
/*  pktinfo	: PktInfo *	: Packet data				*/
/*  (returns)	: bool		: State changed				*/

bool	CapBase::CheckStart
	(	PktInfo	*pktinfo
	)
{
	/* If we are not currently capturing and there is a start	*/
	/* trigger then evaluate it to see if we should now start	*/
	/* capturing.							*/
	if ((state == MSWaiting) && ((mode & TR_START) != 0))
		if (bpf_filter (start.bf_insns,
					(uchar *)pktinfo->data,
					pktinfo->piHdr.pktlen,
					pktinfo->piHdr.caplen))
		{
			state = MSRunning ;
			return	true ;
		}

	return	false	;
}

/*  CapBase								*/
/*  CheckStop	: Check for capture stop on event trigger		*/
/*  pktinfo	: PktInfo *	: Packet data				*/
/*  (returns)	: bool		: State changed				*/

bool	CapBase::CheckStop
	(	PktInfo	*pktinfo
	)
{
	/* This is similar to the start event check, except that we	*/
	/* need to distinguish between repeat and a halt (run until	*/
	/* stop or one-shot).						*/
	if ((state == MSRunning) && ((mode & TR_STOP) != 0))
		if (bpf_filter (stop.bf_insns,
					(uchar *)pktinfo->data,
					pktinfo->piHdr.pktlen,
					pktinfo->piHdr.caplen))
		{
			state	= mode == TR_REPEAT ? MSWaiting : MSHalted ;
			return	true ;
		}

	return	false	;
}

/*  CapBase								*/
/*  CheckAccept	: Check for acceptable packet				*/
/*  pktinfo	: PktInfo *	: Packet data				*/
/*  (returns)	: bool		: Packet accepted			*/

bool	CapBase::CheckAccept
	(	PktInfo	*pktinfo
	)
{
	return	(state == MSRunning) &&
		bpf_filter (program.bf_insns,
					(uchar *)pktinfo->data,
					pktinfo->piHdr.pktlen,
					pktinfo->piHdr.caplen) ;
}

/*  CapBase								*/
/*  CheckEvent	: Check for event					*/
/*  pktinfo	: PktInfo *	: Packet data				*/
/*  eidx	: int		: Event index				*/
/*  (returns)	: bool		: Packet accepted			*/

bool	CapBase::CheckEvent
	(	PktInfo	*pktinfo,
		int	eidx
	)
{
	return	(state == MSRunning) &&
		(events[eidx].set  ) &&
		bpf_filter (events[eidx].event.bf_insns,
					(uchar *)pktinfo->data,
					pktinfo->piHdr.pktlen,
					pktinfo->piHdr.caplen) ;
}

/*  CapBase								*/
/*  Build	: Build capture information				*/
/*  iface	: cchar *	: Interface name			*/
/*  snaplen	: int		: Packet capture length			*/
/*  triginf	: TrigInf *	: Trigger and filter information	*/
/*  eventinf	: EventInf *	: Event information			*/
/*  prefilter	: cchar *	: User pre-filter			*/
/*  (returns)	: bool		: Success				*/

bool	CapBase::Build
	(	cchar	 *iface,
		int	 snaplen,
		TrigInf	 *triginf,
		EventInf *eventinf,
		cchar	 *prefilter
	)
{
	bool	error	;
	BPFI32	netmask	;
	pcap_t	*pd	;

	/* Get an connection to the interface. The error flag is set	*/
	/* and an error message returned if there is an error.		*/
	if (netpcap != NULL)
	{	drop_NetPCap (this, netpcap) ;
		netpcap	= NULL ;
	}

	netpcap	= get_NetPCap (this, iface, snaplen, prefilter, error, ebuf) ;
	if (error) return false ;


	netmask	= netpcap->getNMask ()	;
	pd	= netpcap->getPCap  ()  ;

	/* Compile the start and stop trigger filters if they have been	*/
	/* specified. Note that we are actually going outside the	*/
	/* libpcap API here, so beware! We also note the trigger mode.	*/
	if (triginf->start != NULL)
		if (!CapCompile (pd, &start, triginf->start, 0, netmask))
		{	strcpy	(ebuf, pcap_geterr (pd)) ;
			return	false	;
		}

		else	PrintProg ("Sstart ", triginf->start, start) ;

	if (triginf->stop  != NULL)
		if (!CapCompile (pd, &stop,  triginf->stop,  0, netmask))
		{	strcpy	(ebuf, pcap_geterr (pd)) ;
			return	false	;
		}
		else	PrintProg ("Sstop  ", triginf->stop,  stop ) ;

	mode	= triginf->mode ;
	minev	= NEVENT ;
	maxev	= -1	 ;

	/* Next compile any events that have been set. The "min/maxev"	*/
	/* values are set so that we do the minimum work in the packet	*/
	/* capture callback.						*/
	for (int eidx = 0 ; eidx < NEVENT ; eidx += 1)
		if (StrIsSet (eventinf->event[eidx].event))
		{
			if (!CapCompile (pd, &events[eidx].event, eventinf->event[eidx].event, 0, netmask))
			{	strcpy	(ebuf, pcap_geterr (pd)) ;
				return	false	;
			}

			if (eidx < minev) minev = eidx ;
			if (eidx > maxev) maxev = eidx ;

			events[eidx].set = true  ;
		}
		else	events[eidx].set = false ;


	/* We also compile the monitor program supplied by the		*/
	/* configuration object.					*/
	if (!CapCompile (pd, &program, triginf->filter, 0, netmask))
	{	strcpy	(ebuf, pcap_geterr (pd)) ;
		return	false	;
	}
	else	PrintProg ("Sfilter", triginf->filter, program) ;


	/* All done successfully, so we can finally return a true	*/
	/* result.							*/
	return	true	;
}

/*  CapBase								*/
/*  SendPktData	: Set packet data					*/
/*  pktinfo	: PktInfo *	: Packet data				*/
/*  tag		: int		: Data structure tag			*/
/*  (returns)	: void		:					*/

void	CapBase::SendPktData
	(	PktInfo	*pktinfo,
		int	tag
	)
{
	/* Set the specified tag in the data structure and then send	*/
	/* it off to the logical capture object.			*/
	pktinfo->tag = tag   ;
	gotPacket (pktinfo)  ;
}

/*  CapBase								*/
/*  Callback	: Monitor process callback function			*/
/*  pktinfo	: PktInfo *	: Packet data				*/
/*  (returns)	: void		:					*/

void	CapBase::Callback
	(	PktInfo	*pktinfo
	)
{
	/* Note that all filtering and trigger trsting is done at this	*/
	/* level inside the capture object. This means that it can be	*/
	/* used both directly from ksnuffle and in a remote sniffer;	*/
	/* also, in the case of a remote sniffer, we do not send any	*/
	/* unneccessary packets.					*/

#ifdef	_SHOWPKT
	fprintf	(stderr, "Callback: mode=%d state=%d\n", mode, state) ;
	fflush	(stderr) ;
#endif

	/* Check for start of capture in case we are running with	*/
	/* triggers enabled.						*/
	if (CheckStart (pktinfo))
		gotState (state) ;

	/* Process the packet if we are capturing and then only if it	*/
	/* is accepted by the main filter program.			*/
	if (CheckAccept (pktinfo))
		SendPktData (pktinfo, RS_PACKET) ;

	/* Check for any events, and notify the main ksnuffle program.	*/
	/* The loop is constructed to do the minimum work, and will	*/
	/* not execute at all if there are no events at all.		*/
	for (int eidx = minev ; eidx <= maxev ; eidx += 1)
		if (CheckEvent (pktinfo, eidx))
			SendPktData (pktinfo, RS_EVENT_0 + eidx) ;

	/* Similarly, check if we should stop capturing as a result of	*/
	/* a trigger.							*/
	if (CheckStop (pktinfo))
		gotState (state) ;
}

/*  CapBase								*/
/*  Execute	: Start capture execution				*/
/*  (returns)	: void		:					*/

bool	CapBase::Execute ()
{
	/* Set the capture state and the displayed state according to	*/
	/* the trigger mode. Note that we should never get the		*/
	/* default; this is an internal error.				*/
	switch (mode)
	{
		case TR_ALL 	:
		case TR_TOSTOP	:
			state  = MSRunning ;
			break	;

		case TR_FROMSTART :
		case TR_ONESHOT	  :
		case TR_REPEAT	  :
			state  = MSWaiting ;
			break	;

		default	:
			state  = MSRunning ;
			break	;
	}

	netpcap->execute (true) ;
	return	true ;
}

/*  CapBase								*/
/*  Finish	: Finish capturing					*/
/*  (returns)	: void		:					*/

void	CapBase::Finish ()
{
	if (netpcap != NULL) netpcap->execute (false) ;
}

/*  CapBase								*/
/*  getNMask	: Get netork mask					*/
/*  (returns)	: BPFI32	: Network mask				*/

BPFI32	CapBase::getNMask ()
{
	return	netpcap->getNMask () ;
}

/*  CapBase								*/
/*  getLType	: Get network interface type				*/
/*  (returns)	: int		: Interface type			*/

int	CapBase::getLType ()
{
	return	netpcap->getLType () ;
}


