/***************************************************************************
                          knd_load.cpp  -  description                              
                             -------------------                                         
    begin                : Thu Mar 25 14:26:46 GMT 1999
                                           
    copyright            : (C) 1999 by Mike Richardson                         
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   * 
 *                                                                         *
 ***************************************************************************/


#include	<stdarg.h>
#include	<qtooltip.h>
#include	<qpainter.h>

#include	"knd_load.h"

/*  setTip	: Format and set a tool tip				*/
/*  w		: QWidget *	: Widget on which to set tool tip	*/
/*  t		: char *	: Tip format text			*/
/*  ...		: ...		: Arguments				*/
/*  (returns)	: void		:					*/

static	void	setTip
	(	QWidget	*w,
		char	*t,
		...
	)
{
	char	buff[80] ;
	va_list aptr	 ;

	va_start (aptr, t) ;
	vsprintf (buff, t, aptr) ;
	QToolTip::remove (w) ;
	QToolTip::add	 (w, buff) ;
}

/*  checkBounds	: Check value against bounds				*/
/*  value	: int		: Value to be checked			*/
/*  upper	: int &		: Current/modified upper bound		*/
/*  lower	: int &		: Current/modified lower bound		*/
/*  minval	: int		: Minimum scale				*/
/*  (returns)	: bool		: TRUE if bounds changed.		*/

bool	checkBounds
	(	int	value,
		int	&upper,
		int	&lower,
		int	minval
	)
{
	int	p10	;

	assert	((value >= 0) && (value < MAXBPS)) ;
	assert	((upper >= 0) && (upper < MAXBPS)) ; 
	assert	((lower >= 0) && (lower < MAXBPS)) ;

	/* See if the value has moved outside the current scale range,	*/
	/* in which case readjust the scale appropriately, keeping to	*/
	/* 1..2..5..10.. intervals.					*/
	if ( ( value  > upper) ||
	     ((value <= lower) && (lower > minval)) )
	{
		for (upper = p10 = minval ;; p10 *= 10)
		{
			lower	= upper   ;
			upper  = p10	   ;
			if (upper > value) break ;

			lower	= upper   ;
			upper	= p10 * 2  ;
			if (upper > value) break ;

			lower	= upper   ;
			upper	= p10 * 5  ;
			if (upper > value) break ;
		}

		return	true	;
	}

	return	false	;
}

/*  KNDLoadHist								*/
/*  KNDLoadHist	: Constructor for load histogram ovbject		*/
/*  parent	: QWidget *	: Parent				*/
/*  nsecs	: int		: Time constant				*/
/*  maxwid	: int		: Maximum width				*/
/*  adjust	: bool		: Automatically adjust scale		*/
/*  (returns)	: KNDLoadHist	:					*/

KNDLoadHist::KNDLoadHist
	(	QWidget	*parent,
		int	nsecs,
		int	maxwid,
		bool	adjust
	)
	:
	QFrame	(parent),
	nsecs	(nsecs),
	maxwid	(maxwid),
	adjust	(adjust)
{
	setLineWidth      (1) ;
	setFrameStyle     (QFrame::Box|QFrame::Plain) ;
	setBackgroundMode (NoBackground) ;

	offsetp	= 0	   ;
	offsetd	= 0	   ;
	insert	= maxwid -1;
	scpref	= MINLOAD  ;
	sccurr	= MINLOAD-1;
	scdown	= MINLOAD  ;
	maxval	= MINLOAD  ;
	tickno	= nsecs    ;
	frozen	= false	   ;
	pktcnt	= new LoadPair[maxwid] ;
	dispval	= new LoadPair[maxwid] ;

	for (int slot = 0 ; slot < maxwid ; slot += 1)
	{	pktcnt [slot].pktlen = 0  ;
		pktcnt [slot].datlen = 0  ;
		dispval[slot].pktlen = -1 ;
		dispval[slot].datlen = -1 ;
	}

	repaint	() ;
}

/*  KNDLoadHist								*/
/*  ~KNDLoadHist: Destructor for load histogram object			*/
/*  (returns)	:		:					*/

KNDLoadHist::~KNDLoadHist ()
{
	if (pktcnt  != NULL) delete pktcnt  ;
	if (dispval != NULL) delete dispval ;
}

/*  KNDLoadHist								*/
/*  addPacket	: Add latest packet information				*/
/*  pktinfo	: PktInfo *	  : Packet information			*/
/*  (returns)	: void		  :					*/

void	KNDLoadHist::addPacket
	(	PktInfo	*pktinfo
	)
{
	int	slot	;

	if (pktinfo->ago < maxwid)
	{
		if ((slot = insert + maxwid - pktinfo->ago) >= maxwid)
				slot -= maxwid ;
		assert ((slot >= 0) && (slot < maxwid)) ;

		/* Add in the packet and data lengths. At this stage,	*/
		/* the upper and lower display bounds remain unchanged	*/
		/* as they are set on the next timer tick when not	*/
		/* frozen						*/
		pktcnt[slot].pktlen += pktinfo->piHdr.pktlen ;
		pktcnt[slot].datlen += pktinfo->datlen ;
	}
}

/*  KNDLoadHist								*/
/*  reScale	: Rescale display bounds				*/
/*  (returns)	: void		:					*/

void	KNDLoadHist::reScale ()
{
	for (int slot = 0 ; slot < maxwid ; slot += 1)
	{	dispval[slot].pktlen = boundary(pktcnt[slot].pktlen) ;
		dispval[slot].datlen = boundary(pktcnt[slot].datlen) ;
	}

	setTip	(this, "Scale = %d bytes/sec", sccurr) ;
}

/*  KNDLoadHist								*/
/*  timerTick	: Handle timer tick					*/
/*  _second	: long		: Second				*/
/*  (returns)	: int		: Current scale				*/

int	KNDLoadHist::timerTick
	(	long	_second
	)
{
	int	posn	;
	int	slot	;
	int	from	;

	if ((from = offsetd + maxwid - width()) >= maxwid) from -= maxwid ;

	second	= _second ;

	/* Decrement the tick count. If it falls to zero then reset it	*/
	/* and process the tick, otherwise return. This gives us an	*/
	/* n-seconds time constant.					*/
	if ((tickno -= 1) > 0) return scpref ;
	tickno = nsecs ;

	/* Scan the displayed part of the loading table to find the	*/
	/* maximum byte count. We could do better if we kept track of	*/
	/* it ...							*/
	posn	= 0	;
	slot	= from	;
	maxval	= -1	; 

	while (posn < width())
	{	assert (pktcnt[slot].pktlen < MAXBPS) ;
		if (pktcnt[slot].pktlen > maxval) maxval = pktcnt[slot].pktlen ;
		posn += 1 ;
		if ((slot += 1) >= maxwid) slot = 0 ;
	}

	/* The next chunk of code relates to adjusting the upper and	*/
	/* lower display points. If we are frozen then these are not	*/
	/* altered, so skip it.						*/
	if (!frozen)
	{
		/* See if the maximum value has moved outside the	*/
		/* current scale range, in which case readjust the	*/
		/* scale and then recalculate all display bounds.	*/
		if (checkBounds (maxval, scpref, scdown, MINLOAD))
		{
			if (adjust)
			{	sccurr	= scpref ;
				reScale ()	 ;
			}
		}
		else
		{	/* No change in scale so calculate the display	*/
			/* bounds for the timer interval that has just	*/
			/* completed.					*/
			if ((slot = offsetp + maxwid - 1) >= maxwid)
				slot -= maxwid ;
			dispval[slot].pktlen = boundary (pktcnt[slot].pktlen) ;
			dispval[slot].datlen = boundary (pktcnt[slot].datlen) ;
		}

		dispval[offsetd].pktlen  = height() ;
		dispval[offsetd].datlen  = height() ;
	}

	/* The loading slot which is currently the first displayed will	*/
	/* be reused, so reset the values, then advance the offsets by	*/
	/* one. If we are not frozen then the display offset follows	*/
	/* this.							*/
	pktcnt[offsetp].pktlen = 0 ;
	pktcnt[offsetp].datlen = 0 ;

	if ((offsetp += 1) >= maxwid) offsetp = 0 ;
	if ((insert  += 1) >= maxwid) insert  = 0 ;
	if (!frozen) offsetd = offsetp ;

	return	scpref	;
}

/*  KNDLoadHist								*/
/*  setScale	: Specifically set the scale				*/
/*  scale	: int		: New scale value			*/
/*  (returns)	: void		:					*/

void	KNDLoadHist::setScale
	(	int	scale
	)
{
	if (sccurr != scale)
	{	sccurr	= scale ;
		reScale	()	;
	}
}

/*  KNDLoadHist								*/
/*  drawContents: Handle redrawing display				*/
/*  painter	: QPainter *	: Paint context				*/
/*  (returns)	: void		:					*/

void	KNDLoadHist::drawContents
	(	QPainter	*painter
	)
{
	int	posn	;
	int	slot	;
	int	from	;

	if ((from = offsetd + maxwid - width()) >= maxwid) from -= maxwid ;

	/* The painting is done in three passes of the loading array,	*/
	/* to paint the upper gray area, the middle red (network	*/
	/* traffic overhead) area, and the lower blue (network traffic	*/
	/* data) area. Doing it this way means that we only need to set	*/
	/* the pen three times, rather than (width*3) times.		*/
	posn	= 0	 	;
	slot	= from		;
	painter->setPen (gray)	;

	while (posn < width())
	{
		painter->drawLine (posn, 0,
				   posn, dispval[slot].pktlen) ;
		posn += 1 ;
		if ((slot += 1) >= maxwid) slot = 0 ;
	}


	posn	= 0		;
	slot	= from		;
	painter->setPen (QColor (255, 0, 0)) ;

	while (posn < width())
	{
		painter->drawLine (posn, dispval[slot].pktlen,
				   posn, dispval[slot].datlen) ;
		posn += 1 ;
		if ((slot += 1) >= maxwid) slot = 0 ;
	}

	posn	= 0		;
	slot	= from		;
	painter->setPen (QColor (0, 0, 255)) ;

	while (posn < maxwid)
	{
		painter->drawLine (posn, dispval[slot].datlen,
				   posn, height()) ;
		posn += 1 ;
		if ((slot += 1) >= maxwid) slot = 0 ;
	}
}

/*  KNDLoadHost								*/
/*  mouseDoubleClickEvent	: Handle mouse double click		*/
/*  e		: QMouseEvent *	: Mouse event				*/
/*  (returns)	:		:					*/

void	KNDLoadHist::mouseDoubleClickEvent
	(	QMouseEvent	*e
	)
{
	emit doubleClick () ;
}

/*  KNDLoadHist								*/
/*  freeze	: Set frozen or unfrozen				*/
/*  _frozen	: bool		: New state				*/
/*  (returns)	: void		:					*/

void	KNDLoadHist::freeze
	(	bool	_frozen
	)
{
	/* If we are becoming unfrozen then we need to calculate	*/
	/* display bounds for all samples, since those that arrived	*/
	/* while we were frozen will not have been processed. This is	*/
	/* true whether the scale changes or not.			*/
	if (!(frozen = _frozen))
	{
		checkBounds (maxval, scpref, scdown, MINLOAD) ;

		if (adjust)
		{	sccurr	= scpref ;
			reScale	() ;
		}

		repaint	    () ;
	}
}

/*  KNDLoadHist								*/
/*  resizeEvent	: Handle resize event					*/
/*  e		: QResizeEvent *: Resize event details			*/
/*  (returns)	: void		:					*/

void	KNDLoadHist::resizeEvent
	(	QResizeEvent	*e
	)
{
	reScale	() ;
//	for (int slot = 0 ; slot < maxwid ; slot += 1)
//	{
//		if (dispval[slot].pktlen == -1) dispval[slot].pktlen = height() ;
//		if (dispval[slot].datlen == -1) dispval[slot].datlen = height() ;
//	}
}


/*  KNDLoadBar								*/
/*  KNDLoadBar	: Constructor for load bar object			*/
/*  parent	: QWidget *	: Parent				*/
/*  nsec	: int		: Time constant				*/
/*  (returns)	: KNDLoadHist	:					*/


KNDLoadBar::KNDLoadBar
	(	QWidget	*parent,
		int	nsecs
	)
	:
	QFrame	(parent),
	nsecs	(nsecs)
{
	setLineWidth      (1) ;
	setFrameStyle     (QFrame::Box|QFrame::Plain) ;
	setBackgroundMode (NoBackground) ;

	offset	= 0	  ;
	insert	= nsecs-1 ;
	scale	= MINLOAD ;
	frozen	= false   ;

	pktcnt	= new LoadPair[nsecs] ;
	memset	(pktcnt, 0, nsecs * sizeof(LoadPair)) ;
	dispval.pktlen	= 0 ;
	dispval.datlen	= 0 ;
	pktlen		= 0 ;
	datlen		= 0 ;
	repaint	()  ;
}

/*  KNDLoadBar								*/
/*  ~KNDLoadBar	: Destructor for load bar object			*/
/*  (returns)	:		:					*/

KNDLoadBar::~KNDLoadBar ()
{
	if (pktcnt != NULL) delete pktcnt ;
}

/*  KNDLoadBar								*/
/*  addPacket	: Add latest packet information				*/
/*  pktinfo	: PktDta *	  : Packet data				*/
/*  (returns)	: void		  :					*/

void	KNDLoadBar::addPacket
	(	PktInfo	*pktinfo
	)
{
	int	slot	;

	if (pktinfo->ago < nsecs)
	{
		if ((slot = insert + nsecs - pktinfo->ago) >= nsecs)
			slot -= nsecs ;
		assert ((slot >= 0) && (slot < nsecs)) ;

		/* Add in the packet and data lengths. The entire	*/
		/* array will be summed when displayed.			*/
		pktcnt[insert].pktlen += pktinfo->piHdr.pktlen ;
		pktcnt[insert].datlen += pktinfo->datlen ;
	}
}

/*  KNDLoadBar								*/
/*  reScale	: Rescale display bounds				*/
/*  (returns)	: void		:					*/

void	KNDLoadBar::reScale ()
{
	dispval.pktlen	 = boundary (pktlen) ;
	dispval.datlen	 = boundary (datlen) ;
}

/*  KNDLoadBar								*/
/*  setScale	: Set the bar scale					*/
/*  s		: int		: New scale				*/
/*  (returns)	: void		:					*/

void	KNDLoadBar::setScale
	(	int	s
	)
{
	if (scale != s)
	{	scale = s  ;
		setTip	(this, "Scale = %d bytes/sec", scale) ;
	}

	reScale () ;
}

/*  KNDLoadBar								*/
/*  timerTick	: Handle timer tick					*/
/*  _second	: long		: Second				*/
/*  (returns)	: void		:					*/

void	KNDLoadBar::timerTick
	(	long	_second
	)
{
	int	slot	;

	second	= _second ;

	/* Sum the packet and data lengths, and then calcuate the load	*/
	/* averages (total traffic and total data).			*/
	for (slot = pktlen = datlen = 0 ; slot < nsecs ; slot += 1)
	{	pktlen	+= pktcnt[slot].pktlen ;
		datlen	+= pktcnt[slot].datlen ;
	}

	pktlen	/= nsecs ;
	datlen	/= nsecs ;

	/* The loading slot which is currently the earliest will be	*/
	/* reused, so reset the values, then advance the offset by one.	*/
	pktcnt[offset].pktlen = 0 ;
	pktcnt[offset].datlen = 0 ;

	if ((offset += 1) >= nsecs) offset = 0 ;
	if ((insert += 1) >= nsecs) insert = 0 ;

	/* NOTE: No need to calculate the bar boundaries as the		*/
	/* KNDGraphic parent object will call KNDLoadBar::setScale	*/
	/* before the bar is redrawn.					*/
}

/*  KNDLoadBar								*/
/*  drawContents: Handle redrawing display				*/
/*  painter	: QPainter *	: Paint context				*/
/*  (returns)	: void		:					*/

void	KNDLoadBar::drawContents
	(	QPainter	*painter
	)
{
	painter->fillRect (0,     	   0, width(), dispval.pktlen, gray) ;
	painter->fillRect (0, dispval.pktlen, width(), dispval.datlen, red ) ;
	painter->fillRect (0, dispval.datlen, width(),       height(), blue) ;
}

/*  KNDLoadBar								*/
/*  mouseDoubleClickEvent	: Handle mouse double click		*/
/*  e		: QMouseEvent *	: Mouse event				*/
/*  (returns)	:		:					*/

void	KNDLoadBar::mouseDoubleClickEvent
	(	QMouseEvent	*e
	)
{
	emit doubleClick () ;
}

/*  KNDLoadBar								*/
/*  freeze	: Set frozen or unfrozen				*/
/*  _frozen	: bool		: New state				*/
/*  (returns)	: void		:					*/

void	KNDLoadBar::freeze
	(	bool	_frozen
	)
{
	if (!(frozen = _frozen))
	{	reScale () ;
		repaint () ;
	}
}



/*  KNDPktHist								*/
/*  KNDPktHist	: Constructor for packet histogram ovbject		*/
/*  parent	: QWidget *	: Parent				*/
/*  nsecs	: int		: Time constant				*/
/*  maxwid	: int		: Maximum width				*/
/*  adjust	: bool		: Automatically adjust scale		*/
/*  (returns)	: KNDPktHist	:					*/

KNDPktHist::KNDPktHist
	(	QWidget	*parent,
		int	nsecs,
		int	maxwid,
		bool	adjust
	)
	:
	QFrame	(parent),
	nsecs	(nsecs),
	maxwid	(maxwid),
	adjust	(adjust)
{
	setLineWidth      (1) ;
	setFrameStyle     (QFrame::Box|QFrame::Plain) ;
	setBackgroundMode (NoBackground) ;

	offsetp	= 0	   ;
	offsetd	= 0	   ;
	insert	= maxwid -1;
	scpref	= MINPCNT  ;
	sccurr	= MINPCNT-1;
	scdown	= MINPCNT  ;
	maxval	= MINPCNT  ;
	tickno	= nsecs    ;
	frozen	= false	   ;
	pktcnt	= new int[maxwid] ;
	dispval	= new int[maxwid] ;

	for (int slot = 0 ; slot < maxwid ; slot += 1)
	{	pktcnt [slot]	= 0  ;
		dispval[slot]	= -1 ;
	}

	repaint	() ;
}

/*  KNDPktHist								*/
/*  ~KNDPktHist: Destructor for load histogram object			*/
/*  (returns)	:		:					*/

KNDPktHist::~KNDPktHist ()
{
	if (pktcnt  != NULL) delete pktcnt  ;
	if (dispval != NULL) delete dispval ;
}

/*  KNDPktHist								*/
/*  addPacket	: Add latest packet information				*/
/*  pktinfo	: PktInfo *	  : Packet information			*/
/*  (returns)	: void		  :					*/

void	KNDPktHist::addPacket
	(	PktInfo		*pktinfo
	)
{
	int	slot	;

	if (pktinfo->ago < maxwid)
	{
		if ((slot = insert + maxwid - pktinfo->ago) >= maxwid)
			slot -= maxwid ;
		assert ((slot >= 0) && (slot < maxwid)) ;

		/* Increment. At this stage, the upper and lower	*/
		/* display bounds remain unchanged - they are set on	*/
		/* the next timer tick when not frozen			*/
		pktcnt[slot] += 1 ;
	}
}

/*  KNDPktHist								*/
/*  reScale	: Rescale display bounds				*/
/*  (returns)	: void		:					*/

void	KNDPktHist::reScale ()
{
	for (int slot = 0 ; slot < maxwid ; slot += 1)
		dispval[slot] = boundary(pktcnt[slot]) ;

	setTip	(this, "Scale = %d packets/sec", sccurr) ;
}

/*  KNDPktHist								*/
/*  timerTick	: Handle timer tick					*/
/*  _second	: long		: Second				*/
/*  (returns)	: int		: Current scale				*/

int	KNDPktHist::timerTick
	(	long	_second
	)
{
	int	posn	  ;
	int	slot	  ;
	int	from	  ;

	if ((from = offsetd + maxwid - width()) >= maxwid) from -= maxwid ;

	second	= _second ;

	/* Decrement the tick count. If it falls to zero then reset it	*/
	/* and process the tick, otherwise return. This gives us an	*/
	/* n-seconds time constant.					*/
	if ((tickno -= 1) > 0) return scpref ;
	tickno = nsecs ;

	/* Scan the displayed part of the loading table to find the	*/
	/* maximum byte count. We could do better if we kept track of	*/
	/* it ...							*/
	posn	= 0	;
	slot	= from	;
	maxval	= -1	;

	while (posn < width())
	{	if (pktcnt[slot] > maxval) maxval = pktcnt[slot] ;
		posn += 1 ;
		if ((slot += 1) >= maxwid) slot = 0 ;
	}

	/* The next chunk of code relates to adjusting the display	*/
	/* value. If we are frozen then these are not altered, so skip	*/
	/* it.								*/
	if (!frozen)
	{
		/* See if the maximum value has moved outside the	*/
		/* current scale range, in which case readjust the	*/
		/* scale and then recalculate all display bounds.	*/
		if (checkBounds (maxval, scpref, scdown, MINPCNT))
		{
			if (adjust)
			{	sccurr	= scpref ;
				reScale ()	 ;
			}
		}
		else
		{	/* No change in scale so calculate the display	*/
			/* bounds for the timer interval that has just	*/
			/* completed.					*/
			if ((slot = offsetp + maxwid - 1) >= maxwid)
				slot -= maxwid ;
			dispval[slot] = boundary (pktcnt[slot]) ;
		}

		dispval[offsetd]  = height() ;
	}

	/* The loading slot which is currently the first displayed will	*/
	/* be reused, so reset the values, then advance the offset by	*/
	/* one. If we are not frozen then the display offset follows	*/
	/* this.							*/
	pktcnt[offsetp] = 0 ;

	if ((offsetp += 1) >= maxwid) offsetp = 0 ;
	if ((insert  += 1) >= maxwid) insert  = 0 ;
	if (!frozen) offsetd = offsetp ;

	return	scpref	;
}

/*  KNDPktHist								*/
/*  setScale	: Specifically set the scale				*/
/*  scale	: int		: New scale value			*/
/*  (returns)	: void		:					*/

void	KNDPktHist::setScale
	(	int	scale
	)
{
	if (sccurr != scale)
	{	sccurr	= scale ;
		reScale	()	;
	}
}

/*  KNDPktHist								*/
/*  drawContents: Handle redrawing display				*/
/*  painter	: QPainter *	: Paint context				*/
/*  (returns)	: void		:					*/

void	KNDPktHist::drawContents
	(	QPainter	*painter
	)
{
	int	posn	;
	int	slot	;
	int	from	;

	if ((from = offsetd + maxwid - width()) >= maxwid) from -= maxwid ;

	/* The painting is done in two passes of the loading array, to	*/
	/* paint the upper gray area, and the lower blue (network	*/
	/* traffic data) area. Doing it this way means that we only	*/
	/* need to set the pen twice, rather than (width2) times.	*/
	posn	= 0	 	;
	slot	= from		;
	painter->setPen (gray)	;

	while (posn < width())
	{
		painter->drawLine (posn, 0, posn, dispval[slot]) ;
		posn += 1 ;
		if ((slot += 1) >= maxwid) slot = 0 ;
	}


	posn	= 0		;
	slot	= from		;
	painter->setPen (QColor (0, 0, 255)) ;

	while (posn < width())
	{
		painter->drawLine (posn, dispval[slot], posn, height()) ;
		posn += 1 ;
		if ((slot += 1) >= maxwid) slot = 0 ;
	}
}

/*  KNDLoadHost								*/
/*  mouseDoubleClickEvent	: Handle mouse double click		*/
/*  e		: QMouseEvent *	: Mouse event				*/
/*  (returns)	:		:					*/

void	KNDPktHist::mouseDoubleClickEvent
	(	QMouseEvent	*e
	)
{
	emit doubleClick () ;
}

/*  KNDPktHist								*/
/*  freeze	: Set frozen or unfrozen				*/
/*  _frozen	: bool		: New state				*/
/*  (returns)	: void		:					*/

void	KNDPktHist::freeze
	(	bool	_frozen
	)
{
	/* If we are becoming unfrozen then we need to calculate	*/
	/* display bounds for all samples, since those that arrived	*/
	/* while we were frozen will not have been processed. This is	*/
	/* true whether the scale changes or not.			*/
	if (!(frozen = _frozen))
	{
		checkBounds (maxval, scpref, scdown, MINPCNT) ;

		if (adjust)
		{	sccurr	= scpref ;
			reScale	() ;
		}

		repaint	    () ;
	}
}

/*  PktHist								*/
/*  resizeEvent	: Handle resize event					*/
/*  e		: QResizeEvent *: Resize event details			*/
/*  (returns)	: void		:					*/

void	KNDPktHist::resizeEvent
	(	QResizeEvent	*e
	)
{
	reScale	() ;

//	for (int slot = 0 ; slot < maxwid ; slot += 1)
//		if (dispval[slot] == -1) dispval[slot] = height() ;
}


/*  KNDPktBar								*/
/*  KNDPktBar	: Constructor for load bar object			*/
/*  parent	: QWidget *	: Parent				*/
/*  nsec	: int		: Time constant				*/
/*  (returns)	: KNDLoadHist	:					*/


KNDPktBar::KNDPktBar
	(	QWidget	*parent,
		int	nsecs
	)
	:
	QFrame	(parent),
	nsecs	(nsecs)
{
	setLineWidth      (1) ;
	setFrameStyle     (QFrame::Box|QFrame::Plain) ;
	setBackgroundMode (NoBackground) ;

	offset	= 0	  ;
	insert	= nsecs-1 ;
	scale	= MINLOAD ;
	frozen	= false   ;

	pktcnt	= new int[nsecs] ;
	memset	(pktcnt, 0, nsecs * sizeof(int)) ;
	dispval	= 0 ;
	repaint	()  ;
}

/*  KNDPktBar								*/
/*  ~KNDPktBar	: Destructor for load bar object			*/
/*  (returns)	:		:					*/

KNDPktBar::~KNDPktBar ()
{
	if (pktcnt != NULL) delete pktcnt ;
}

/*  KNDPktBar								*/
/*  addPacket	: Add latest packet information				*/
/*  pktinfo	: PktDta *	  : Packet data				*/
/*  (returns)	: void		  :					*/

void	KNDPktBar::addPacket
	(	PktInfo	*pktinfo
	)
{
	int	slot	;

	if (pktinfo->ago < nsecs)
	{
		if ((slot = insert + nsecs - pktinfo->ago) >= nsecs)
			slot -= nsecs ;
		assert ((slot >= 0) && (slot < nsecs)) ;

		/* Increment the packet count. The array will be summed	*/
		/* when displayed.					*/
		pktcnt[insert] += 1 ;
	}
}

/*  KNDPktBar								*/
/*  reScale	: Rescale display bounds				*/
/*  (returns)	: void		:					*/

void	KNDPktBar::reScale ()
{
	int	slot	;
	int	pktlen	;

	/* Sum the packet and data lengths, and calculate the upper and	*/
	/* lower bar boundaries. We do this here rather than when the	*/
	/* bar is redrawn, as typically "timerTick" is called before	*/
	/* "drawContents", and we would otherwise loose one sample.	*/
	for (slot = pktlen = 0 ; slot < nsecs ; slot += 1)
		pktlen	+= pktcnt[slot] ;

	dispval	= boundary (pktlen / nsecs) ;
}

/*  KNDPktBar								*/
/*  setScale	: Set the bar scale					*/
/*  s		: int		: New scale				*/
/*  (returns)	: void		:					*/

void	KNDPktBar::setScale
	(	int	s
	)
{
	if (scale != s)
	{	scale = s  ;
		setTip	(this, "Scale = %d packets/sec", scale) ;
	}

	reScale () ;
}

/*  KNDPktBar								*/
/*  timerTick	: Handle timer tick					*/
/*  _second	: long		: Second				*/
/*  (returns)	: void		:					*/

void	KNDPktBar::timerTick
	(	long	_second
	)
{
	second	= _second ;

	/* The loading slot which is currently the earliest will be	*/
	/* reused, so reset the values, then advance the offset by one.	*/
	pktcnt[offset] = 0 ;

	if ((offset += 1) >= nsecs) offset = 0 ;
	if ((insert += 1) >= nsecs) insert = 0 ;

	/* In principal, we should call "reScale" here to recalculate	*/
	/* the bar boundary. However, KNDGraphic operates such that	*/
	/* "setScale" is called immediately before a repaint is forced,	*/
	/* so we save the effort.					*/
}

/*  KNDPktBar								*/
/*  drawContents: Handle redrawing display				*/
/*  painter	: QPainter *	: Paint context				*/
/*  (returns)	: void		:					*/

void	KNDPktBar::drawContents
	(	QPainter	*painter
	)
{
	painter->fillRect (0,       0, width(),  dispval, gray) ;
	painter->fillRect (0, dispval, width(), height(), blue) ;
}

/*  KNDPktBar								*/
/*  mouseDoubleClickEvent	: Handle mouse double click		*/
/*  e		: QMouseEvent *	: Mouse event				*/
/*  (returns)	:		:					*/

void	KNDPktBar::mouseDoubleClickEvent
	(	QMouseEvent	*e
	)
{
	emit doubleClick () ;
}

/*  KNDPktBar								*/
/*  freeze	: Set frozen or unfrozen				*/
/*  _frozen	: bool		: New state				*/
/*  (returns)	: void		:					*/

void	KNDPktBar::freeze
	(	bool	_frozen
	)
{
	if (!(frozen = _frozen))
	{	reScale () ;
		repaint () ;
	}
}



