/***************************************************************************
                          netpcap.cpp  -  description                              
                             -------------------                                         
    begin                : Thu Mar 25 14:26:46 GMT 1999
                                           
    copyright            : (C) 1999,2000 by Mike Richardson                         
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   * 
 *                                                                         *
 ***************************************************************************/

#include	<errno.h>
#include	<unistd.h>
#include	<fcntl.h>

#include	"ksnuffle.h"
#include	"pcaphdr.h"
#include	"netpcap.h"
#include	"capture.h"


/*  pcapCallback: Monitor process callback function			*/
/*  user	: u_char *	: User data (actually NetPCap *)	*/
/*  h		: pcap_pkthdr *	: Packet header information		*/
/*  p		: u_char *	: Packet data				*/
/*  (returns)	: void		:					*/

static	void	pcapCallback
	(	u_char		*user,
		const	struct	pcap_pkthdr *h,
		const	u_char	*p
	)
{
	/* Dive back into the actual capture object so that we can	*/
	/* cleanly access the member variables.				*/
	((NetPCap *)user)->Callback (h, p) ;
}

/*  NetPCap								*/
/*  NetPCap	: Constructor for low-level capture class		*/
/*  iface	: cchar *	: Interface name			*/
/*  snaplen	: int		: Capture length			*/
/*  prefilter	: cchar *	: User packet pre-filter		*/
/*  error	: bool &	: Error flag				*/
/*  ebuf	: char *	: Error message buffer			*/
/*  (returns)	: NetPCap	:					*/

NetPCap::NetPCap
	(	cchar	*iface,
		int	snaplen,
		cchar	*prefilter,
		bool	&error,
		char	*ebuf
	)
	:
	iface	(strdup (iface)),
	snaplen	(snaplen)
{
	BPFI32	locnet	;

	caplist	= NULL	;

	fprintf	(stderr, "NetPCap: create for %s,%d\n", iface, snaplen) ;

	/* First stage is to open the interface, which we do in		*/
	/* promiscuous mode ....					*/
	if ((pd = pcap_open_live ((char *)iface, snaplen, 1, 1000, ebuf)) == NULL)
	{	error	= true	;
		return	;
	}

	/* Second stage is to determin the local network address and	*/
	/* the network mask for the interface ....			*/
	if (pcap_lookupnet ((char *)iface, &locnet, &netmask, ebuf) != 0)
	{	error	= true	;
		return	;
	}

	/* Compile the user-prefilter program; note that this will be	*/
	/* null if there are no restrictions. Then set this as the	*/
	/* filter.							*/
	if (!CapCompile (pd, &preflt, prefilter == NULL ? "" : prefilter, 0, netmask))
	{	strcpy	(ebuf, pcap_geterr (pd)) ;
		error	= true	;
		return	;
	}

	if (pcap_setfilter (pd, &preflt) < 0)
	{	strcpy	(ebuf, pcap_geterr (pd)) ;
		error	= true	;
		return	;
	}

	error	= false	;
}

/*  NetPCap								*/
/*  ~NetPCap	: Destructor for low-level capture class		*/
/*  (returns)	:		:					*/

NetPCap::~NetPCap ()
{
	fprintf	(stderr, "NetPCap: destroy: %s\n", iface) ;

	if (pd != NULL)
	{	pcap_close (pd) ;
		pd  = NULL ;
	}

	free	(iface) ;
}

/*  NetPCap								*/
/*  addCapture	: Add upper-level capture object			*/
/*  capture	: CapBase *	: Object to add				*/
/*  (returns)	:		:					*/

void	NetPCap::addCapture
	(	CapBase	*capture
	)
{
	capture->next	= caplist ;
	caplist		= capture ;
}

/*  NetPCap								*/
/*  delCapture	: Remove upper-level capture object			*/
/*  capture	: CapBase *	: Object to remove			*/
/*  (returns)	: bool		: True if no longer in use		*/

bool	NetPCap::delCapture
	(	CapBase	*capture
	)
{
	for (CapBase **lvc = &caplist ; *lvc != NULL ; lvc = &(*lvc)->next)
		if (*lvc == capture)
		{	*lvc	= capture->next ;
			capture->next = NULL	;
			break	;
		}

	return	caplist == NULL ;
}

/*  NetPCap								*/
/*  sockRead	: Handle incoming captured data				*/
/*  ksock	: KSocket *	: Capture socket			*/
/*  (returns)	: void		:					*/

void	NetPCap::sockRead
	(	int	sock
	)
{
	/* Call "pcap_read" to do the work. This should read a packet	*/
	/* from the socket, do some processing, and then call the	*/
	/* callback routine, which gets back to this object at the	*/
	/* ::Callback method. This way means that although pcap_read is	*/
	/* not defined as part of the libpcap interface, we do not need	*/
	/* to actually reimplement any of libpcap.			*/
	pcap_read (pd, -1, pcapCallback, (u_char *)this) ; 
}

/*  NetPCap								*/
/*  Callback	: packet capture callback routine			*/
/*  h		: pcap_pkthdr *	: Capture header			*/
/*  p		: u_char *	: Captured data				*/
/*  (returns)	: void		:					*/

void	NetPCap::Callback
	(	const	struct	pcap_pkthdr *h,
		const	u_char	*p
	)
{
	/* Run through the list of all attached upper-level capture	*/
	/* objects, passing the packet to each of them in turn.		*/
	PktInfo			*pktinfo	= getPktInfo () ;

#ifdef	_DEBUG
	fprintf	(stderr, "NeTPCap::Callback: got packet\n") ;
#endif	/* _DEBUG */

	pktinfo->piHdr.pktlen	= h->len		;
	pktinfo->piHdr.caplen	= h->caplen		;
	pktinfo->piHdr.tval	= h->ts			;
	pktinfo->piHdr.accept	= pd->md.stat.ps_recv   ;
	pktinfo->piHdr.dropped	= pd->md.stat.ps_drop   ;
	pktinfo->piHdr.missed	= pd->md.stat.ps_ifdrop ;
	memcpy	(pktinfo->data, p, h->caplen < sizeof(pktinfo->data) ? h->caplen : sizeof(pktinfo->data)) ;

	for (CapBase *cap = caplist ; cap != NULL ; cap = cap->next)
		cap->Callback (pktinfo) ;

	freePktInfo (pktinfo) ;
}

/*  NetPCap								*/
/*  execute	: Start/stop execution					*/
/*  exec	: bool		: True to start				*/
/*  (returns)	: void		:					*/

void	NetPCap::execute
	(	bool	exec
	)
{
}

/*  NetPCap								*/
/*  suitable	: See if object can be used for new sniffer		*/
/*  _iface	: cchar *	: Requested interface			*/
/*  _snaplen	: int		: Requested capture snap length		*/
/*  (returns)	: bool		: True if suitable			*/

bool	NetPCap::suitable
	(	cchar	*_iface,
		int	_snaplen
	)
{
	return	(iface == iface) && (snaplen >= snaplen) ;
}


static	NetPCap	*capList	;


/*  get_NetPCap	: Get a low-level capture object			*/
/*  capture	: CapBase *	: Upper-level capture parent		*/
/*  iface	: cchar *	: Interface name			*/
/*  snaplen	: int		: Capture length			*/
/*  prefilter	: cchar *	: User packet pre-filter		*/
/*  error	: bool *	: Error flag				*/
/*  ebuf	: char *	: Error message buffer			*/
/*  (returns)	: NetPCap *	: New object				*/

NetPCap	*get_NetPCap
	(	CapBase	*capture,
		cchar	*iface,
		int	snaplen,
		cchar	*prefilter,
		bool	&error,
		char	*ebuf
	)
{
	/* First look to see if we have an existing connection to the	*/
	/* interface; if so then increment its use count and attatch	*/
	/* the caller to it.						*/
	NetPCap	*cap	;

	for (cap = capList ; cap != NULL ; cap = cap->next)
		if (cap->suitable (iface, snaplen))
		{	cap->addCapture (capture) ;
			error	= false ;
			return	cap	;
		}

	/* No usable extant connection, so we need to create a new one,	*/
	/* to which we attach the calling sniffer.			*/
	cap		= new NetPCap (iface, snaplen, prefilter, error, ebuf) ;
	cap->next	= capList ;
	capList		= cap	  ;
	cap->addCapture (capture) ;
	return	cap	;
}


/*  drop_NetPCap: Drop connection					*/
/*  capture	: CapBase *	: Upper-level capture object		*/
/*  netpcap	: NetPCap *	: Low-level capture being dropped	*/
/*  (returns)	: void		:					*/

void	drop_NetPCap
	(	CapBase	*capture,
		NetPCap	*netpcap
	)
{
	if (netpcap->delCapture (capture))
	{
		for (NetPCap **lvc = &capList ; *lvc != NULL ; lvc = &(*lvc)->next)
			if (*lvc == netpcap)
			{	*lvc	= netpcap->next	;
				delete	netpcap		;
				netpcap	= NULL		;
				break	;
			}

		if (netpcap != NULL)
			fprintf	(stderr, "ksnuffle: drop_NetPCap: lost\n") ;
	}
}
